const { Container, DependencyInjector } = require('flitter-di')
const { test_connect } = require('../common')
const { expect } = require('chai')
const { MongoClient, ObjectId } = require('mongodb')
const uuid = require('uuid/v4')

describe('limit/sort/filter and chaining', function() {
    const Connection = require('../../src/services/Connection')
    const Model = require('../../src/model/Model')
    const di = new DependencyInjector(new Container({ scaffold: Connection }))
    class TestModel extends Model {
        static collection = uuid()

        static get schema() {
            return {
                name: String,
                number: Number,
            }
        }
    }
    di.make(TestModel)

    before(async function() {
        await di.service('scaffold').open(test_connect)
        for ( let i = 1; i < 15; i++ ) {
            await (new TestModel({number: i, name: `number ${i}`})).save()
        }
    })

    it('should allow limiting', async function() {
        const normal = await TestModel.find()
        const limit_2 = await TestModel.limit(2).find()
        const limit_7 = await TestModel.limit(7).find()
        expect(normal.length).to.be.equal(14)
        expect(limit_2.length).to.be.equal(2)
        expect(limit_7.length).to.be.equal(7)
    })

    it('should allow sorting', async function() {
        const sorted = await TestModel.sort('-number').find()
        const rev_sorted = await TestModel.sort('+number').find()
        expect(sorted.map(x => x.number)).to.be.eql([14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1])
        expect(rev_sorted.map(x => x.number)).to.be.eql([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14])
    })

    it('should allow filtering', async function() {
        const filter = await TestModel.filter()
        filter.less_than('number', 8).greater_than('number', 5)

        const results = await filter.end().find()
        const result_nums = results.map(x => x.number)
        expect(result_nums.length).to.be.equal(2)
        expect(result_nums).to.include(7)
        expect(result_nums).to.include(6)
    })

    it('should allow chaining limit, sort, and filter', async function() {
        const filter = await TestModel.filter()
        const results = await filter.greater_than('number', 3).end().limit(2).sort('+num').find()
        expect(results.length).to.be.equal(2)
        expect(results.map(x => x.number)).to.be.eql([4, 5])
    })

    after(function() {
        di.service('scaffold').close()
    })
})
