const { expect } = require('chai')
const sinon = require('sinon')
const { ObjectId } = require('mongodb')

const ModelProxy = require('../../src/proxy/model/ModelProxy')

describe('the model proxies', function() {
    describe('the base model proxy class', function() {
        let mp
        let fake_ref
        let fake_bldr
        let fake_resultset
        let fake_cursor
        let fake_coll
        beforeEach(function() {
            fake_resultset = [
                { _id: new ObjectId, delete: sinon.spy() },
                { _id: new ObjectId, delete: sinon.spy() },
                { _id: new ObjectId, delete: sinon.spy() },
            ]
            fake_cursor = {
                toArray: function() {
                    return fake_resultset
                }
            }
            fake_cursor.limit = sinon.fake.returns(fake_cursor)
            fake_cursor.sort = sinon.fake.returns(fake_cursor)
            fake_coll = {
                deleteMany: sinon.spy(),
            }
            fake_ref = {
                cursor: sinon.fake.returns(fake_cursor),
                from_cursor: sinon.fake.returns(fake_resultset),
                __collection: sinon.fake.returns(fake_coll),
                filter: sinon.fake.returns('a fake filter')
            }
            fake_bldr = {
                apply: sinon.fake(x => x),
            }

            mp = new ModelProxy(fake_ref, fake_bldr)
        })

        it('should be a functional class', function() {
            expect(ModelProxy).to.be.a(`function`)
            expect(new ModelProxy()).to.be.an.instanceof(ModelProxy)
        })

        it('should expect one argument on construction', function() {
            expect(ModelProxy.length).to.be.equal(1)
        })

        it('should set the builder and reference members on construct', function() {
            const ref = {}
            const bldr = {}
            const mp1 = new ModelProxy(ref)
            const mp2 = new ModelProxy(ref, bldr)

            expect(mp1.builder).to.exist
            expect(mp1.reference).to.be.equal(ref)
            expect(mp2.builder).to.be.equal(bldr)
            expect(mp2.reference).to.be.equal(ref)
        })

        describe('the __collection proxy', function() {
            it('should return the __collection on the reference', async function() {
                const fake_coll = {}
                const ref = {
                    __collection: sinon.fake.returns(fake_coll)
                }
                const mp = new ModelProxy(ref)

                const ret = mp.__collection()
                expect(ret).to.be.a('promise')

                const val = await ret
                expect(val).to.be.equal(fake_coll)
                expect(ref.__collection.calledOnce).to.be.true
            })
        })

        describe('the cursor proxy', function() {
            it('should be a function', function() {
                expect(mp.cursor).to.be.a('function')
            })

            it('should call the cursor on the reference', async function() {
                const a = {}
                const b = {}
                await mp.cursor(a, b)
                expect(fake_ref.cursor.calledOnce).to.be.true
                expect(fake_ref.cursor.getCall(0).args[0]).to.be.equal(a)
                expect(fake_ref.cursor.getCall(0).args[1]).to.be.equal(b)
            })

            it('should return the cursor after the builder has modified it', async function() {
                const result = await mp.cursor()
                expect(fake_bldr.apply.calledOnce).to.be.true
                expect(fake_bldr.apply.getCall(0).args[0]).to.be.equal(fake_cursor)
                expect(result).to.be.equal(fake_cursor)
            })
        })

        describe('the find proxy', function() {
            it('should be a function', function() {
                expect(mp.find).to.be.a('function')
            })

            it('should fetch a cursor with the provided filter and opts', async function() {
                const a = {}
                const b = {}
                await mp.find(a, b)
                expect(fake_ref.cursor.calledOnce).to.be.true
                expect(fake_ref.cursor.getCall(0).args).to.be.eql([a, b])
            })

            it('should return the reference results from the cursor', async function() {
                const ret = await mp.find()
                expect(fake_ref.from_cursor.calledOnce).to.be.true
                expect(fake_ref.from_cursor.getCall(0).args[0]).to.be.equal(fake_cursor)
                expect(ret).to.be.eql(fake_resultset)
            })
        })

        describe('the findOne proxy', function() {
            it('should be a function', function() {
                expect(mp.findOne).to.be.a('function')
            })

            it('should fetch a cursor with the provided filter and opts', async function() {
                const a = {}
                const b = {}
                await mp.findOne(a, b)
                expect(fake_ref.cursor.calledOnce).to.be.true
                expect(fake_ref.cursor.getCall(0).args).to.be.eql([a, b])
            })

            it('should return the first reference result from the cursor', async function() {
                const ret = await mp.findOne()
                expect(fake_ref.from_cursor.calledOnce).to.be.true
                expect(fake_ref.from_cursor.getCall(0).args[0]).to.be.equal(fake_cursor)
                expect(ret).to.be.equal(fake_resultset[0])
            })
        })

        describe('the deleteMany proxy', function() {
            it('should be a function', function() {
                expect(mp.deleteMany).to.be.a('function')
            })

            it('should call deleteMany on the collection for the ids matched by the cursor', async function() {
                const arg = {_id: {$in: fake_resultset.map(x => x._id)}}
                await mp.deleteMany()
                expect(fake_coll.deleteMany.calledOnce).to.be.true
                expect(fake_coll.deleteMany.getCall(0).args[0]).to.be.eql(arg)
            })
        })

        describe('the deleteOne proxy', function() {
            it('should be a function', function() {
                expect(mp.deleteOne).to.be.a('function')
            })

            it('should fetch a cursor with the provided filter and opts', async function() {
                const a = {}
                const b = {}
                await mp.deleteOne(a, b)
                expect(fake_ref.cursor.calledOnce).to.be.true
                expect(fake_ref.cursor.getCall(0).args).to.be.eql([a, b])
            })

            it('should delete the first reference result from the cursor', async function() {
                await mp.deleteOne()
                expect(fake_resultset[0].delete.calledOnce).to.be.true
                expect(fake_resultset[1].delete.calledOnce).to.be.false
                expect(fake_resultset[2].delete.calledOnce).to.be.false
            })
        })

        describe('the from_cursor proxy', function() {
            it('should be a function', function() {
                expect(mp.from_cursor).to.be.a('function')
            })

            it('should return a promise', function() {
                expect(mp.from_cursor(fake_cursor)).to.be.a('promise')
            })

            it('should resolve to the reference result set from the cursor', async function() {
                const ret = await mp.from_cursor(fake_cursor)
                expect(fake_ref.from_cursor.calledOnce).to.be.true
                expect(ret).to.be.equal(fake_ref.from_cursor(fake_cursor))
            })
        })

        describe('the limit proxy', function() {
            it('should be a function', function() {
                expect(mp.limit).to.be.a('function')
            })

            it('should expect one argument', function() {
                expect(mp.limit.length).to.be.equal(1)
            })

            it('should return a ModelProxy', function() {
                expect(mp.limit(1)).to.be.an.instanceof(ModelProxy)
            })

            it('should limit the resulting cursor', async function() {
                const n = 2
                const ret_mp = mp.limit(n)
                const ret_cursor = await ret_mp.cursor()
                expect(ret_cursor.limit.calledOnce).to.be.true
                expect(ret_cursor.limit.getCall(0).args[0]).to.be.equal(n)
            })
        })

        describe('the sort proxy', function() {
            it('should be a function', function() {
                expect(mp.sort).to.be.a('function')
            })

            it('should return a ModelProxy', function() {
                expect(mp.sort('test_field')).to.be.an.instanceof(ModelProxy)
            })

            it('should sort the resulting cursor', async function() {
                const field = '-fake_field'
                const expect_obj = { fake_field: -1 }
                const ret_mp = mp.sort(field)
                const ret_cursor = await ret_mp.cursor()
                expect(ret_cursor.sort.calledOnce).to.be.true
                expect(ret_cursor.sort.getCall(0).args[0]).to.be.eql(expect_obj)
            })
        })

        describe('the filter proxy', function() {
            it('should return a promise', function() {
                expect(mp.filter()).to.be.a('promise')
            })

            it('should return the filter for the reference', async function() {
                const val = await mp.filter()
                expect(fake_ref.filter.calledOnce).to.be.true
                expect(fake_ref.filter.getCall(0).args[0]).to.be.equal(mp)
                expect(val).to.be.equal('a fake filter')
            })

            it('should allow overriding the filter ref', async function() {
                const a = {}
                const val = await mp.filter(a)
                expect(fake_ref.filter.calledOnce).to.be.true
                expect(fake_ref.filter.getCall(0).args[0]).to.be.equal(a)
                expect(val).to.be.equal('a fake filter')
            })
        })
    })

    describe('the sort proxy class', function() {
        const SortProxy = require('../../src/proxy/model/SortProxy')
        it('should be a function', function() {
            expect(SortProxy).to.be.a('function')
        })

        it('should expect two params on construction', function() {
            expect(SortProxy.length).to.be.equal(2)
        })

        it('should extend ModelProxy', function() {
            expect(SortProxy.prototype).to.be.an.instanceof(ModelProxy)
        })

        it('should call the builder sort method', function() {
            const fake_model = {}
            const fake_bldr = {
                sort: sinon.spy()
            }
            const sorts = ['-fake_field']
            const sp = new SortProxy(fake_model, sorts, fake_bldr)

            expect(fake_bldr.sort.calledOnce).to.be.true
            expect(fake_bldr.sort.getCall(0).args[0]).to.be.eql(sorts)
            expect(sp).to.be.an.instanceof(SortProxy)
        })

        it('should store the sorts values', function() {
            const sorts = new SortProxy({}, ['-a', 'test', '+sort'])
            expect(sorts._sorts).to.be.eql(['-a', 'test', '+sort'])
        })
    })

    describe('the limit proxy class', function() {
        const LimitProxy = require('../../src/proxy/model/LimitProxy')
        it('should be a function', function() {
            expect(LimitProxy).to.be.a('function')
        })

        it('should expect two params on construction', function() {
            expect(LimitProxy.length).to.be.equal(2)
        })

        it('should extend ModelProxy', function() {
            expect(LimitProxy.prototype).to.be.an.instanceof(ModelProxy)
        })

        it('should call the builder limit method', function() {
            const fake_model = {}
            const fake_bldr = {
                limit: sinon.spy()
            }
            const limit = 5
            const lp = new LimitProxy(fake_model, limit, fake_bldr)

            expect(fake_bldr.limit.calledOnce).to.be.true
            expect(fake_bldr.limit.getCall(0).args[0]).to.be.equal(limit)
            expect(lp).to.be.an.instanceof(LimitProxy)
        })

        it('should store the limit value', function() {
            const limit = new LimitProxy({}, 448)
            expect(limit._limit).to.be.equal(448)
        })
    })

    describe('the filter proxy class', function() {
        const FilterProxy = require('../../src/proxy/model/FilterProxy')
        it('should be a function', function() {
            expect(FilterProxy).to.be.a('function')
        })

        it('should expect two params on construction', function() {
            expect(FilterProxy.length).to.be.equal(2)
        })

        it('should extend ModelProxy', function() {
            expect(FilterProxy.prototype).to.be.an.instanceof(ModelProxy)
        })

        it('should call the builder filter method', function() {
            const fake_model = {}
            const fake_bldr = {
                filter: sinon.spy()
            }
            const filter = {}
            const fp = new FilterProxy(fake_model, filter, fake_bldr)

            expect(fake_bldr.filter.calledOnce).to.be.true
            expect(fake_bldr.filter.getCall(0).args[0]).to.be.equal(filter)
            expect(fp).to.be.an.instanceof(FilterProxy)
        })
    })
})
